# -*- coding: utf-8 -*-

__all__ = ['_utils','functions','main','stats']
dependencies = ['pandas','numpy']
missing_dependencies = []

for dependency in dependencies:
    try:
        __import__(dependency)
    except ImportError as e:
        missing_dependencies.append(dependency)

if missing_dependencies:
    raise ImportError("Missing required dependencies {0}".format(missing_dependencies))
del(dependencies)
del(dependency)
del(missing_dependencies)

from mr_clean.core.api import *



# -*- coding: utf-8 -*-
import pandas as _pd

def normalize(df, style = 'mean'):
    """ Returns a normalized version of a DataFrame or Series
    Parameters:
    df - DataFrame or Series
        The data to normalize
    style - function or string, default 'mean'
        The style to use when computing the norms. Takes 'mean' or 'minmax' to
        do mean or min-max normalization respectively. User-defined functions that take
        a pandas Series as input and return a normalized pandas Series are also accepted
    """
    if style == 'mean':
        df_mean,df_std = df.mean(),df.std()
        return (df-df_mean)/df_std
    elif style == 'minmax':
        col_min,col_max = df.min(),df.max()
        return (df-col_min)/(col_max-col_min)
    else:
        return style(df)

def col_normalize(df,col_names, style = 'mean'):
    """ Returns a normalized version of a set of columns in the input Dataframe 
    Parameters:
    df - pandas DataFrame
        The input data to normalize
    col_names - list or string, default None
        The column(s) to use when computing the norms
    style - function or string, default 'mean'
        The style to use when computing the norms. Takes 'mean' or 'minmax' to
        do mean or min-max normalization respectively. User-defined functions that take
        a pandas Series as input and return a normalized pandas Series are also accepted
    """
    return normalize(df[col_names],style)

def row_normalize(df,row_names,style = 'mean'):
    """ Returns a normalized version of a set of rows in the input Dataframe
    Parameters:
    df - pandas DataFrame
        The input data to normalize
    row_names - list or string, default None
        The row(s) to use when computing the norms
    style - function or string, default 'mean'
        The style to use when computing the norms. Takes 'mean' or 'minmax' to
        do mean or min-max normalization respectively. User-defined functions that take
        a pandas Series as input and return a normalized pandas Series are also accepted
    """
    return normalize(df.loc[row_names,:],style)

def norms(df, col_names = None,row_names = None,style = 'mean', as_group = False, axis = 0):
    """ Returns a normalized version of the input Dataframe
    Parameters:
    df - pandas DataFrame
        The input data to normalize
    col_names - list or string, default None
        The column(s) to use when computing the norms
    row_names - list or string, default None
        The row(s) to use when computing the norms
    style - function or string, default 'mean'
        The style to use when computing the norms. Takes 'mean' or 'minmax' to
        do mean or min-max normalization respectively. User-defined functions that take
        a pandas Series as input and return a normalized pandas Series are also accepted
    as_group - bool, default False
        Whether to normalize accross the entire range or by row/column. If true, will normalize
        by the entire DataFrame
    axis - int or string, default 0
        Which axis to perform the normalization on. Accepts 0 or 'columns' to do normalization by
        columns, and 1 or 'rows' to do normalization by rows
    """
    if col_names is None:
        if row_names is not None:
            df = df.loc[row_names,:]
    else:
        if row_names is None:
            df = df.loc[:,col_names]
        else:
            df = df.loc[row_names,col_names]
    if as_group:
        return normalize(df,style)
    if axis == 0 or str(axis).startswith('column'):
        return _pd.concat([col_normalize(df,col_name,style) for col_name in df.columns],axis = 1)
    elif axis == 1 or str(axis).startswith('row'):
        return _pd.concat([row_normalize(df,row_name,style) for row_name in df.index])
    else:
        return normalize(df,style)




# -*- coding: utf-8 -*-
import anytree

# Resources
# https://github.com/cceh/suffix-tree
# https://github.com/Rerito/suffix-tree/blob/master/suffixtree.h
# http://brenden.github.io/ukkonen-animation/

def suffix_tree(str_list):
    # Build suffix tree using Ukkonen’s algorithm
    
    # Return at end
    pass

def suffix_tree_recurse(str_list):
    # Recursively build suffix tree using memoization
    
    # Return at end
    pass


# Maybe extend anytree.Node?


class SNode(anytree.Node):
    
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)
        self.end = False
        self.index_dict = None
    
    def add_index(self,index):
        if self.end:
            self.index_dict[index]=True
        else:
            self.index_dict = {index:True}
            self.end = True
    
    def remove_index(self,index):
        try:
            if self.index_dict[index]:
                self.index_dict[index] = False
                return
        except:
            pass
        print("index wasn't in node's index list, couldn't perform remove_index of {}".format(index))
    
    # String index stored as child node whose value is of type int
    # Index Child node only exists when end == True


# -*- coding: utf-8 -*-
import pandas as _pd
import mr_clean._utils.data_handling as _dutils
import mr_clean._utils.globals as _globals
from collections import deque

#None of these functions should have much logic.


# --------- Blind ops on entire table ---------------------

def colname_gen(df,col_name = 'unnamed_col'):
    """ Returns a column name that isn't in the specified DataFrame
    Parameters:
    df - DataFrame
        DataFrame to analyze
    col_name - string, default 'unnamed_col'
        Column name to use as the base value for the generated column name
    """
    if col_name not in df.keys():
        yield col_name
    id_number = 0
    while True:
        col_name = col_name + str(id_number)
        if col_name in df.keys():
            id_number+=1
        else:
            return col_name

def clean_colnames(df):
    """ Cleans the column names on a DataFrame and returns the new names
    Parameters:
    df - DataFrame
        The DataFrame to clean
    """
    col_list = []
    for index in range(_dutils.cols(df)):
        col_list.append(df.columns[index].strip().lower().replace(' ','_'))
    return col_list

# ----- Blind ops on single columns -------------

def col_strip(df,col_name,dest = False):
    """ Performs str.strip() a column of a DataFrame
    Parameters:
    df - DataFrame
        DataFrame to operate on
    col_name - string
        Name of column to strip
    dest - bool, default False
        Whether to apply the result to the DataFrame or return it.
        True is apply, False is return.
    """
    if dest:
        df[col_name] = df[col_name].str.strip()
    else:
        return df[col_name].str.strip()

def col_scrubf(df,col_name,which,count = 1,dest = False):
    """ Removes characters from the front of the entries in DataFrame for a column
    Parameters:
    df - DataFrame
        DataFrame to operate on
    col_name - string
        Name of column to scrub
    which - boolean array
        Boolean array that determines which elements to scrub, where True means scrub
    count - integer, default 1
        amount of characters to scrub
    dest - bool, default False
        Whether to apply the result to the DataFrame or return it.
        True is apply, False is return.
    """
    if dest:
        df.loc[which,col_name] = df.loc[which,col_name].str[count:]
    else:
        new_col = df[col_name].copy()
        new_col[which] = df.loc[which,col_name].str[count:]
        return new_col

def col_scrubb(df,col_name,which, count = 1,dest = False):
    """ Removes characters from the back of the entries in DataFrame for a column
    Parameters:
    df - DataFrame
        DataFrame to operate on
    col_name - string
        Name of column to scrub
    which - boolean array
        Boolean array that determines which elements to scrub, where True means scrub
    count - integer, default 1
        amount of characters to scrub
    dest - bool, default False
        Whether to apply the result to the DataFrame or return it.
        True is apply, False is return.
    """
    if dest:
        df.loc[which,col_name] = df.loc[which,col_name].str[:-count]
    else:
        new_col = df[col_name].copy()
        new_col[which] = df.loc[which,col_name].str[:-count]
        return new_col

def col_to_numeric(df,col_name, dest = False):
    """ Coerces a column in a DataFrame to numeric
    Parameters:
    df - DataFrame
        DataFrame to operate on
    col_name - string
        Name of column to coerce
    dest - bool, default False
        Whether to apply the result to the DataFrame or return it.
        True is apply, False is return.
    """
    new_col = _pd.to_numeric(df[col_name], errors = 'coerce')
    if dest:
        set_col(df,col_name,new_col)
    else:
        return new_col

def col_to_dt(df,col_name,set_format = None,infer_format = True, dest = False):
    """ Coerces a column in a DataFrame to datetime
    Parameters:
    df - DataFrame
        DataFrame to operate on
    col_name - string
        Name of column to coerce
    dest - bool, default False
        Whether to apply the result to the DataFrame or return it.
        True is apply, False is return.
    """
    new_col = _pd.to_datetime(df[col_name],errors = 'coerce',
                                    format = set_format,infer_datetime_format = infer_format)
    if dest:
        set_col(df,col_name,new_col)
    else:
        return new_col

def col_to_cat(df,col_name, dest = False):
    """ Coerces a column in a DataFrame to categorical
    Parameters:
    df - DataFrame
        DataFrame to operate on
    col_name - string
        Name of column to coerce
    dest - bool, default False
        Whether to apply the result to the DataFrame or return it.
        True is apply, False is return.
    """
    new_col = df[col_name].astype('category')
    if dest:
        set_col(df,col_name,new_col)
    else:
        return new_col

# ------- Blind ops on single columns (destructive) ------

def set_col(df,col_name, new_values):
    """ Sets a column in a DataFrame
    Parameters:
    df - DataFrame
        DataFrame to operate on
    col_name - string
        Name of column to set
    new_values - Series
        The new values for that column
    """
    df[col_name] = new_values

def col_rename(df,col_name,new_col_name):
    """ Changes a column name in a DataFrame
    Parameters:
    df - DataFrame
        DataFrame to operate on
    col_name - string
        Name of column to change
    new_col_name - string
        New name of column
    """
    col_list = list(df.columns)
    for index,value in enumerate(col_list):
        if value == col_name:
            col_list[index] = new_col_name
            break
    df.columns = col_list

def col_mod(df,col_name,func,*args,**kwargs):
    """ Changes a column of a DataFrame according to a given function
    Parameters:
    df - DataFrame
        DataFrame to operate on
    col_name - string
        Name of column to modify
    func - function
        The function to use to modify the column
    """
    backup = df[col_name].copy()
    try:
        return_val = func(df,col_name,*args,**kwargs)
        if return_val is not None:
            set_col(df,col_name,return_val)
    except:
        df[col_name] = backup

# ------- Blind operations on multiple columns ----------

def cols_strip(df,col_list, dest = False):
    """ Performs str.strip() a column of a DataFrame
    Parameters:
    df - DataFrame
        DataFrame to operate on
    col_list - list of strings
        names of columns to strip
    dest - bool, default False
        Whether to apply the result to the DataFrame or return it.
        True is apply, False is return.
    """
    if not dest:
        return _pd.DataFrame({col_name:col_strip(df,col_name) for col_name in col_list})
    for col_name in col_list:
        col_strip(df,col_name,dest)

def cols_to_numeric(df, col_list,dest = False):
    """ Coerces a list of columns to numeric
    Parameters:
    df - DataFrame
        DataFrame to operate on
    col_list - list of strings
        names of columns to coerce
    dest - bool, default False
        Whether to apply the result to the DataFrame or return it.
        True is apply, False is return.
    """
    if not dest:
        return _pd.DataFrame({col_name:col_to_numeric(df,col_name) for col_name in col_list})
    for col_name in col_list:
        col_to_numeric(df,col_name,dest)

def cols_to_dt(df, col_list,set_format = None,infer_format = True,dest = False):
    """ Coerces a list of columns to datetime
    Parameters:
    df - DataFrame
        DataFrame to operate on
    col_list - list of strings
        names of columns to coerce
    dest - bool, default False
        Whether to apply the result to the DataFrame or return it.
        True is apply, False is return.
    """
    if not dest:
        return _pd.DataFrame({col_name:col_to_dt(df,col_name,set_format,infer_format) for col_name in col_list})
    for col_name in col_list:
        col_to_dt(df,col_name,set_format,infer_format,dest)

def cols_to_cat(df, col_list,dest = False):
    """ Coerces a list of columns to categorical
    Parameters:
    df - DataFrame
        DataFrame to operate on
    col_list - list of strings
        names of columns to coerce
    dest - bool, default False
        Whether to apply the result to the DataFrame or return it.
        True is apply, False is return.
    """
    # Convert a list of columns to categorical
    if not dest:
        return _pd.DataFrame({col_name:col_to_cat(df,col_name) for col_name in col_list})
    for col_name in col_list:
        col_to_cat(df,col_name,dest)

def cols_(df,col_list,func,*args,**kwargs):
    """ Do a function over a list of columns and return the result
    Parameters:
    df - DataFrame
        DataFrame to operate on
    col_list - list of strings
        names of columns to coerce
    func - function
        function to use
    """
    return _pd.DataFrame({col_name:func(df,col_name,*args,**kwargs) for col_name in col_list})

# ------- Blind operations on multiple columns (destructive) ----------

def cols_rename(df,col_names, new_col_names):
    """ Rename a set of columns in a DataFrame
    Parameters:
    df - DataFrame
        DataFrame to operate on
    col_names - list of strings
        names of columns to change
    new_col_names - list of strings
        new names for old columns (order should be same as col_names)
    """
    assert len(col_names) == len(new_col_names)
    for old_name,new_name in zip(col_names,new_col_names):
        col_rename(df,old_name,new_name)

# ------- Get formatting information ----------

def col_dtypes(df): # Does some work to reduce possibility of errors and stuff
    """ Returns dictionary of datatypes in a DataFrame (uses string representation)
    Parameters:
    df - DataFrame
        The DataFrame to return the object types of
    
    Pandas datatypes are as follows:
    object,number,bool,datetime,category,timedelta,datetimetz
    This method uses queues and iterates over the columns in linear time.
    It does extra steps to ensure that no further work with numpy datatypes needs
    to be done.
    """
    test_list = [col_isobj,col_isnum,col_isbool,col_isdt,col_iscat,col_istdelt,col_isdtz]
    deque_list = [(deque(col_method(df)),name) \
                  for col_method,name in zip(test_list,_globals.__dtype_names) if len(col_method(df))]
    type_dict = {}
    for que, name in deque_list:
        while len(que):
            type_dict[que.popleft()] = name
    return type_dict

def col_isobj(df, col_name = None):
    """ Returns a list of columns that are of type object. If col_name is specified, returns 
    whether the column in the DataFrame is of type 'object' instead.
    Parameters:
    df - DataFrame
        DataFrame to check
    col_name - string, default None
        If specified, this function will True if df[col_name] is of type 'object'
    """
    col_list = df.select_dtypes(include = 'object').columns
    if col_name is None:
        return col_list
    else:
        return col_name in col_list

def col_isnum(df,col_name = None):
    """ Returns a list of columns that are of type 'number'. If col_name is specified, returns 
    whether the column in the DataFrame is of type 'number' instead.
    Parameters:
    df - DataFrame
        DataFrame to check
    col_name - string, default None
        If specified, this function will True if df[col_name] is of type 'number'
    """
    col_list = df.select_dtypes(include = 'number').columns
    if col_name is None:
        return col_list
    else:
        return col_name in col_list

def col_isbool(df,col_name = None):
    """ Returns a list of columns that are of type 'bool'. If col_name is specified, returns 
    whether the column in the DataFrame is of type 'bool' instead.
    Parameters:
    df - DataFrame
        DataFrame to check
    col_name - string, default None
        If specified, this function will True if df[col_name] is of type 'bool'
    """
    col_list = df.select_dtypes(include = 'bool').columns
    if col_name is None:
        return col_list
    else:
        return col_name in col_list

def col_isdt(df,col_name = None):
    """ Returns a list of columns that are of type 'datetime'. If col_name is specified, returns 
    whether the column in the DataFrame is of type 'datetime' instead.
    Parameters:
    df - DataFrame
        DataFrame to check
    col_name - string, default None
        If specified, this function will True if df[col_name] is of type 'datetime'
    """
    col_list = df.select_dtypes(include = 'datetime').columns
    if col_name is None:
        return col_list
    else:
        return col_name in col_list

def col_iscat(df,col_name = None):
    """ Returns a list of columns that are of type 'category'. If col_name is specified, returns 
    whether the column in the DataFrame is of type 'category' instead.
    Parameters:
    df - DataFrame
        DataFrame to check
    col_name - string, default None
        If specified, this function will True if df[col_name] is of type 'category'
    """
    col_list = df.select_dtypes(include = 'category').columns
    if col_name is None:
        return col_list
    else:
        return col_name in col_list

def col_istdelt(df,col_name = None):
    """ Returns a list of columns that are of type 'timedelta'. If col_name is specified, returns 
    whether the column in the DataFrame is of type 'timedelta' instead.
    Parameters:
    df - DataFrame
        DataFrame to check
    col_name - string, default None
        If specified, this function will True if df[col_name] is of type 'timedelta'
    """
    col_list = df.select_dtypes(include = 'timedelta').columns
    if col_name is None:
        return col_list
    else:
        return col_name in col_list

def col_isdtz(df,col_name = None):
    """ Returns a list of columns that are of type 'datetimetz'. If col_name is specified, returns 
    whether the column in the DataFrame is of type 'datetimetz' instead.
    Parameters:
    df - DataFrame
        DataFrame to check
    col_name - string, default None
        If specified, this function will True if df[col_name] is of type 'datetimetz'
    """
    col_list = df.select_dtypes(include = 'datetimetz').columns
    if col_name is None:
        return col_list
    else:
        return col_name in col_list



# -*- coding: utf-8 -*-
import pandas as pd
import io
import shutil
import math

# --------- Memory Information ---------------------

# Simple statement of memory usage
def memory_statement(savings,task_name,statement):
    return  statement \
              .format( convert_memory(savings) ,task_name )

# Convert memeory to a readable form
def convert_memory(memory_value):
    unit_list = ['KB','MB','GB','TB']
    index = 0
    memory = memory_value / 1024
    while memory > 1000 and index < 3:
        memory/=1024
        index+=1
    return '{} {}'.format(round( memory, 1),unit_list[index])

# --------- Formatting outputs ---------------------

def preview(df,preview_rows = 20):#,preview_max_cols = 0):
    """ Returns a preview of a dataframe, which contains both header
    rows and tail rows.
    """
    if preview_rows < 4:
        preview_rows = 4
    preview_rows = min(preview_rows,df.shape[0])
    outer = math.floor(preview_rows / 4)
    return pd.concat([df.head(outer),
                      df[outer:-outer].sample(preview_rows-2*outer),
                      df.tail(outer)])

def get_info(df, verbose = None,max_cols = None, memory_usage = None, null_counts = None):
    """ Returns the .info() output of a dataframe
    """
    assert type(df) is pd.DataFrame
    buffer = io.StringIO()
    df.info(verbose, buffer, max_cols, memory_usage, null_counts)
    return buffer.getvalue()

def title_line(text):
    """Returns a string that represents the
    text as a title blurb
    """
    columns = shutil.get_terminal_size()[0]
    start = columns // 2 - len(text) // 2
    output = '='*columns + '\n\n' + \
            ' ' * start + str(text) + "\n\n" + \
            '='*columns + '\n'
    return output

#------------ Outputting --------------------

# Outputs to a file. If and only if ouput_safe is false, it will overwrite existing files
def output_to_file(data,output_file,output_safe):
    try:
        with open(output_file,'x' if output_safe else 'w') as file:
                file.write(data)
    except FileExistsError:
        print("Nothing outputted: file '{}' already exists".format(output_file))

def format_row(row, total_rows):
    output = []
    for item in row:
        output.append( "{} ({}%)".format(item, round(item/total_rows*100,2)) )
    return pd.Series(output)



# -*- coding: utf-8 -*-
import mr_clean._utils.data_handling as _utils
import mr_clean.core.functions.basics as _basics

def smart_scrub(df,col_name,error_rate = 0):
    """ Scrubs from the front and back of an 'object' column in a DataFrame
    until the scrub would semantically alter the contents of the column. If only a 
    subset of the elements in the column are scrubbed, then a boolean array indicating which
    elements have been scrubbed is appended to the dataframe. Returns a tuple of the strings removed
    from the front and back of the elements
    df - DataFrame
        DataFrame to scrub
    col_name - string
        Name of column to scrub
    error_rate - number, default 0
        The maximum amount of values this function can ignore while scrubbing, expressed as a
        fraction of the total amount of rows in the dataframe. 
    """
    scrubf = smart_scrubf(df,col_name,error_rate)
    scrubb = smart_scrubb(df,col_name,error_rate)
    return (scrubf, scrubb)

def smart_scrubf(df,col_name,error_rate = 0):
    """ Scrubs from the front of an 'object' column in a DataFrame
    until the scrub would semantically alter the contents of the column. If only a 
    subset of the elements in the column are scrubbed, then a boolean array indicating which
    elements have been scrubbed is appended to the dataframe. Returns the string that was scrubbed
    df - DataFrame
        DataFrame to scrub
    col_name - string
        Name of column to scrub
    error_rate - number, default 0
        The maximum amount of values this function can ignore while scrubbing, expressed as a
        fraction of the total amount of rows in the dataframe. 
    """
    scrubbed = ""
    while True:
        valcounts = df[col_name].str[:len(scrubbed)+1].value_counts()
        if not len(valcounts):
            break
        if not valcounts[0] >= (1-error_rate) * _utils.rows(df):
            break
        scrubbed=valcounts.index[0]
    if scrubbed == '':
        return None
    which = df[col_name].str.startswith(scrubbed)
    _basics.col_scrubf(df,col_name,which,len(scrubbed),True)
    if not which.all():
        new_col_name = _basics.colname_gen(df,"{}_sf-{}".format(col_name,scrubbed))
        df[new_col_name] = which
    return scrubbed

def smart_scrubb(df,col_name,error_rate = 0):
    """ Scrubs from the back of an 'object' column in a DataFrame
    until the scrub would semantically alter the contents of the column. If only a 
    subset of the elements in the column are scrubbed, then a boolean array indicating which
    elements have been scrubbed is appended to the dataframe. Returns the string that was scrubbed.
    df - DataFrame
        DataFrame to scrub
    col_name - string
        Name of column to scrub
    error_rate - number, default 0
        The maximum amount of values this function can ignore while scrubbing, expressed as a
        fraction of the total amount of rows in the dataframe. 
    """
    scrubbed = ""
    while True:
        valcounts = df[col_name].str[-len(scrubbed)-1:].value_counts()
        if not len(valcounts):
            break
        if not valcounts[0] >= (1-error_rate) * _utils.rows(df):
            break
        scrubbed=valcounts.index[0]
    if scrubbed == '':
        return None
    which = df[col_name].str.endswith(scrubbed)
    _basics.col_scrubb(df,col_name,which,len(scrubbed),True)
    if not which.all():
        new_col_name = _basics.colname_gen(df,"{}_sb-{}".format(col_name,scrubbed))
        df[new_col_name] = which
    return scrubbed



# -*- coding: utf-8 -*-
import re
from mr_clean._utils.substrings.substr_main import substr_list

def smart_colnames(df, cutoff = .5):
    """ Replace spaces and dashes with underscores and make everything lowercase
    remove underscores before and after
    replace multiple underscores with just one
    Create dictionary of taken words, with counts cooresponding to occurence rate

    Use N-grams to do this faster and easier -- dictionary of string tuples? Maybe an ordered list instead?

    Loop:
        Find words/phrases common among large portion of columns (according to cutoff)
        For each word/phrase:
            Replace each word in phrase with a single letter
            Loop:
                If already in dictionary:
                    add another letter to the abbrev. based on the original word phrase
                    try again
                else:
                    break
            replace phrase with abbreviation for all instances.
            Remove words of phrase from dictionary
            add abbreviation to the dictionary
            add (phrase,abbreviation) to list
    """

    col_list = []
    # Removing whitespace and making everything lowercase
    for col_name in df.columns:
		col_name = re.sub('[^a-zA-Z0-9]',' ',col_name)
        col_name = col_name.strip().lower()
        col_name = re.sub('\s+','_',col_name)

        col_list.append(col_name)

    # Dictionary of words in the columns
    word_dict,max_len = col_words(col_list)
    # Max len is 1 greater than the length of largest string tuple that we'll test for
    word_list = list(word_dict.keys())




    return col_list


def col_words(col_list):
    """ Get the list of words in the columns
    Parameters:
    col_list - list of strings
        List of column names
    """
    col_dict = {}
    max_len = 1
    for col_name in col_list:
        words = col_name.split('_')
        max_len = max(len(words),max_len)
        for word in words:
            if word in col_dict.keys():
                col_dict[word]+=1
            else:
                col_dict[word]=1
    return (col_dict,max_len)

def abbrev(phrase,word_length = 1):
    """ Abbreviates a phrase
    Parameters:
    phrase - str
        Phrase to abbreviate
    word_length - int, default 1
        Length of each word after abbreviation

    Ex:
        word_word_2word
    word_length 1: ww2w
    word_length 2: wowo2wo
    """
    word_list = phrase.split('_')
    abbreviation = ''
    for word in word_list:
        recording = 0
        for character in word:
            if isnum(character):
                recording = 0
                abbreviation+=character
            elif recording < word_length:
                abbreviation+=character
                recording+=1
    return abbreviation

def isnum(character):
    return character in ['0','1,','2','3','4','5','6','7','8','9']



# -*- coding: utf-8 -*-

# Utility function get longest common substrings from a list of substrings

import pandas as pd
from anytree import Node
from mr_clean._utils.substrings.suffix_tree import suffix_tree
from mr_clean._utils.substrings.substr_list import substr_list

def substr_list(str_list,length = -1,min_freq = 2,repeats = True):
    """ Gets a list of common substrings, ordered by length
    Parameters:
    str_list - list of strings
        List of strings to find the common substrings of
    length - int, default -1
        Lower limit for the length of the strings in the list. All strings with length
        equal to or less than the limit are returned. If set to -1, then all will be returned
    min_freq - int, default 2
        Minimum frequency of the substring necessary to include it in the list
    repeats - bool, default True
        Whether or not elements in the list that originate as substrings of other elements will
        remain in the list.
        
    WARNING:
    If set to false, runtime of this function could increase substantially. Checking for repeats is 
    currently on the order of O(n^2*s) operation, where n is the amount of elements and s is the
    average length of the strings in the list.
    
    Ex:
        substr_list(['dasdf','dfwed','_as'])
        output = ['df','as','a','s','d','f','w','e','_']
    
    Ex: 
        substr_list(['dasdf','dfwed','_as'],length = 2)
        output = ['df','as']
    Ex:
        substr_list(['dasdf','dfwed','_as'],repeats = False)
        output = ['df','as','w','e','d','_']
        
        substr_list(['wasd','wasd','awa'],repeats = False)
        output = ['wasd','wa']
        Since 'sd','d','as', and 'asd' all are only present as substrings of 'wasd', they
        are not included in the final results. However, 'wa' is present both in 'wasd' and
        in 'awa', outside of the string 'wasd', and thus is included.
    """
    
    str_list = pd.Series(str_list)
    
    # Documentation for tree structure https://pypi.org/project/anytree/
    
    # Method: Suffix Tree https://en.wikipedia.org/wiki/Generalized_suffix_tree
    
    # Description: https://en.wikipedia.org/wiki/Longest_common_substring_problem
    
    # TODO Lol the entire thing
    
    # Plan
    
    # Get suffix tree
    
    # Get list of common substrings
    
    # Filter by length if necessary
    
    # Filter by repeats if necessary
    
    
    return substr_list


def sort_by_length(str_list):
    pass

def filter_by_length(str_list):
    pass




# -*- coding: utf-8 -*-

from mr_clean.core.functions.basics import *
from mr_clean.core.functions.scrub import *
#from mr_clean.core.functions.colnames import *
#from mr_clean.core.functions.coerce import *

from mr_clean.core.stats.summary import *
from mr_clean.core.stats.transform import *
from mr_clean.core.stats.regression import *

#from mr_clean.core.tools.cleaner import *
from mr_clean.core.tools.mrc import clean
from mr_clean.core.tools.summarize import summarize
from mr_clean.core.tools.diagnose import diagnose



# -*- coding: utf-8 -*-
import pandas as pd
import mr_clean._utils.io as _io
import mr_clean._utils.data_handling as _utils
import mr_clean.core.stats.summary as stats

def diagnose(df,preview_rows = 2,
            display_max_cols = 0,display_width = None): 
    """ Prints information about the DataFrame pertinent to data cleaning.

    Parameters
    ----------
    df - DataFrame
        The DataFrame to summarize
    preview_rows - int, default 5
        Amount of rows to preview from the head and tail of the DataFrame
    display_max_cols - int, default None
        Maximum amount of columns to display. If set to None, all columns will be displayed.
        If set to 0, only as many as fit in the screen's width will be displayed
    display_width - int, default None
        Width of output. Can be width of file or width of console for printing.
        Set to None for pandas to detect it from console.
    """
    assert type(df) is pd.DataFrame
    # Diagnose problems with the data formats that can be addressed in cleaning
    
     # Get initial display settings
    initial_max_cols = pd.get_option('display.max_columns')
    initial_max_rows = pd.get_option('display.max_rows')
    initial_width = pd.get_option('display.width')

    # Reformat displays
    pd.set_option('display.max_columns', display_max_cols)
    pd.set_option('display.max_rows',None)
    if display_width is not None:
            pd.set_option('display.width',display_width)

    # --------Values of data-----------
    df_preview = _io.preview(df,preview_rows)
    
    df_info = _io.get_info(df,verbose = True, max_cols = display_max_cols, 
                               memory_usage = 'deep',null_counts = True)
    
    dtypes = stats.dtypes_summary(df).apply(_io.format_row,args = [_utils.rows(df)],axis = 1)
    
    potential_outliers = stats.df_outliers(df).dropna(axis = 1,how = 'all')
    potential_outliers = potential_outliers if _utils.rows(potential_outliers) \
                    else None

    # ----------Build lists------------
    
    title_list = \
    ['Preview','Info',
     'Data Types Summary','Potential Outliers']
    info_list = \
    [df_preview,df_info,
     dtypes,potential_outliers]
    error_list = [None,None,
                  None,'No potential outliers.']
    
    # ----------Build output------------
    
    output = ''
    for title, value,error_text in zip(title_list,info_list,error_list):
        if value is None:
            value = "{} skipped: {}".format(title,error_text)
        if str(value).endswith('\n'):
            value = value[:-1]
        output+='{}\n{}\n\n'.format(_io.title_line(title),value)

    # ----------Send to file/print to console------------
        # Potentially could change this to allow for output_safe to work with directories
    print(output)

    # Reset display settings
    pd.set_option('display.max_columns', initial_max_cols)
    pd.set_option('display.max_rows', initial_max_rows)
    pd.set_option('display.width', initial_width)    
    
    


# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
import mr_clean._utils.data_handling as _utils
import mr_clean.core.functions.basics as _basics

def dtypes_summary(df):
    """ Takes in a dataframe and returns a dataframe with
    information on the data-types present in each column.
    Parameters:
    df - DataFrame
        Dataframe to summarize
    """
    output_df = pd.DataFrame([])
    row_count = df.shape[0]
    row_indexes = ['rows_numerical','rows_string','rows_date_time','category_count','largest_category','rows_na','rows_total']
    for colname in df:
        data = df[colname] # data is the pandas series associated with this column
        # number of numerical values in the column
        rows_numerical = pd.to_numeric(data,errors = 'coerce').count()
        # number of values that can't be coerced to a numerical
        rows_string = row_count - rows_numerical
        # number of values that can be coerced to a date-time object
        rows_date_time = pd.to_datetime(data,errors = 'coerce',infer_datetime_format = True).count()
        # categories in column
        value_counts = data.value_counts().reset_index()
        # number of different values in the dataframe
        categories = len(value_counts)
        # largest category
        largest_category = value_counts.iloc[0,1]
        # number of null/missing values
        rows_na = data.isnull().sum()
        # build the output list
        output_data = [rows_numerical, rows_string, rows_date_time, categories, 
                       largest_category,rows_na,row_count]
        # add to dataframe
        output_df.loc[:,colname] = pd.Series(output_data)

    # row names
    output_df.index = row_indexes
    return output_df

def percentiles(df,q =[0,.1,.2,.3,.4,.5,.6,.7,.8,.9,1]):
    """ Takes a dataframe and returns the quantiles for each column,
    or an error message if there are no columns with quantitative data.
    Parameters:
    df - DataFrame
        The dataframe to analyze
    q - number or list of numbers, default [0,.1,.2,.3,.4,.5,.6,.7,.8,.9,1]
        Quantiles to get
    """
    try:
        return df.quantile(q = q)
    except ValueError:
        return None#"No columns with numeric data."

def df_outliers(df,sensitivity = 1.5):
    """ Finds outliers in the dataframe.
    Parameters:
    df - DataFrame
        The DataFrame to analyze.
    sensitivity - number, default 1.5
        The value to multipy by the iter-quartile range when determining outliers. This number is used
        for categorical data as well.
    """
    outlier_df = df.copy()
    dtypes = _basics.col_dtypes(df)
    for col_name in df.columns:
        outlier_df.loc[~outliers(df[col_name],'bool',dtypes[col_name],sensitivity),col_name] = np.nan
    outlier_df = outlier_df.dropna(how = 'all')
    return outlier_df

def outliers(df,output_type = 'values',dtype = 'number',sensitivity = 1.5):# can output boolean array or values
    """ Returns potential outliers as either a boolean array or a subset of the original.
    Parameters:
    df - array_like
        Series or dataframe to check
    output_type - string, default 'values'
        if 'values' is specified, then will output the values in the series that are suspected
        outliers. Else, a boolean array will be outputted, where True means the value is an outlier
    dtype - string, default 'number'
        the way to treat the object. Possible values are 'number','datetime',
        'timedelt','datetimetz','category',or 'object'
    sensitivity - number, default 1.5
        The value to multipy by the iter-quartile range when determining outliers. This number is used
        for categorical data as well.
    """
    if dtype in ('number','datetime','timedelt','datetimetz'):
        if not dtype == 'number':
            df = pd.to_numeric(df,errors = 'coerce')
        quart25, quart75 = percentiles(df,q = [.25,.75]) 
        out_range= sensitivity * (quart75 - quart25)
        lower_bound,upper_bound = quart25-out_range, quart75+out_range
        bool_array = (df < lower_bound)|(df > upper_bound)
    else:
        value_counts = df.value_counts() # Trying to find categorical outliers.
        quart25 = cum_percentile(value_counts,.25)
        quart75 = cum_percentile(value_counts,.75)
        out_values = int(sensitivity * (quart75 - quart25) + quart75 + 1)
        if out_values >= len(value_counts):
            bool_array = _utils.bc_vec(df,value = False)
        else:
            outlier_values = value_counts[value_counts <= value_counts.iloc[out_values]].index
            bool_array = df.isin(outlier_values)
        
    if output_type == 'values':
        return  df[bool_array]
    return bool_array

def cum_percentile(series,q):
    """ Takes a series of ordered frequencies and returns the value at a specified quantile
    Parameters:
    series - Series
        The series to analyze
    q - number
        Quantile to get the value of
    """
    total = series.sum()
    cum_sum = series.cumsum()
    return sum(cum_sum < total*q)
    



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
#import mr_clean.core.functions.basics as basics
from mr_clean._utils.data_handling import rows,cols

def corr_matrix(df,method = 'pearson'):
    """ Returns a matrix of correlations between columns of a DataFrame. For categorical columns,
    it first changes those to a set of dummy variable columns. Booleans are converted to
    numerical as well. Also ignores any indexes set on the DataFrame
    Parameters:
    df - DataFrame
        DataFrame to analyze
    method - {'pearson', 'kendall', 'spearman'}
        * pearson : standard correlation coefficient
        * kendall : Kendall Tau correlation coefficient
        * spearman : Spearman rank correlation
    """
    # Remove all but categoricals,booleans, and numerics
    df = df.reset_index(drop = True)
    cat_cols = df.select_dtypes(include = 'category')
    bool_cols = df.select_dtypes(include = 'bool')
    df = df.select_dtypes(include = 'number')
    if not cols(df) + cols(bool_cols) + cols(cat_cols):
        return None # quit if there's none of the possible datatypes present

    #Convert categoricals to boolean columns
    insert = np.ones(rows(df))
    for col_name in cat_cols:
        cat_df = pd.concat([cat_cols[[col_name]],pd.Series(insert)],axis = 1) # Add a column of ones as values for the pivot
        cat_ptable = cat_df.pivot(columns = col_name).reset_index(drop = True)
        cat_ptable.columns = [col_name+ "_{}".format(value) for value in
                                  cat_ptable.columns.get_level_values(col_name)]
        df = pd.concat([df,cat_ptable.fillna(0)],axis = 1)
    df = pd.concat([df,bool_cols * 1], axis = 1)
    return df.corr(method,0)

#
#%%%
#np.array([1,2,3,1,2,3,1,2,3,2])
#df = pd.DataFrame({'col1':np.array([1,2,3,1,2,3,1,2,3,2]),'col2':np.array([1,2,3,1,2,3,1,2,3,2]),'col3':np.array([1,2,3,1,2,3,1,2,3,2])})
#df.index = ['oneboi','twodne','threeoain','fouroqwn','five','size','sven','ight','noef','tinnny']
#col_name = 'col1'
#df[col_name] = df[col_name].astype('category')
#df['col2'] = df['col2'].astype('category')
#ndf = df.pivot(columns = col_name).reset_index(drop = True).iloc[:,0:len(df[col_name].cat.categories)]
#ndf.columns = [col_name+ "_{}".format(value) for value in ndf.columns.get_le
#vel_values(col_name)]
#fdf = ndf.notna() * 1


# -*- coding: utf-8 -*-
import pandas as pd
import mr_clean._utils.io as _io
import os
import mr_clean._utils.data_handling as _utils
import mr_clean.core.stats.summary as stats
import mr_clean.core.stats.regression as regstats
# Pre-cleaning

# This method takes in a DataFrame object, as well as a few parameters,
# and outputs a DataFrame that summarizes some of the possible problems
# that might have to be addressed in cleaning
def summarize(df,preview_rows = 8,
            display_max_cols = None,display_width = None,
            output_path = None, output_safe = True,to_folder = False): 
    """ Prints information about the DataFrame to a file or to the prompt.

    Parameters
    ----------
    df - DataFrame
        The DataFrame to summarize
    preview_rows - int, default 5
        Amount of rows to preview from the head and tail of the DataFrame
    display_max_cols - int, default None
        Maximum amount of columns to display. If set to None, all columns will be displayed.
        If set to 0, only as many as fit in the screen's width will be displayed
    display_width - int, default None
        Width of output. Can be width of file or width of console for printing.
        Set to None for pandas to detect it from console.
    output_path - path-like, default None
        If not None, this will be used as the path of the output file, and this
        function will print to a file instead of to the prompt
    output_safe - boolean, default True
        If True and output_file is not None, this function will not overwrite any
        existing files.
    output_csv: boolean, default False
        If True, will output to a directory with name of output_path with all data in 
        csv format. WARNING: If set to true, this function will overwrite existing files
        in the directory with the following names: 
            ['Preview.csv','Describe.csv','Info.csv','Percentile Details.csv',
             'Missing Values Summary.csv','Potential Outliers.csv','Correlation Matrix.csv']
    """
    assert type(df) is pd.DataFrame
    
    # Reformat displays
    initial_settings = pd_settings(display_max_cols, None, display_width)

    # --------Values of data-----------
    df_preview = _io.preview(df,preview_rows)
    df_desc_num, df_desc_cat = detailed_desc(df)
    percent_values = stats.percentiles(df)
    potential_outliers = stats.df_outliers(df).dropna(axis = 1,how = 'all')
    potential_outliers = potential_outliers if _utils.rows(potential_outliers) else None
    corr_values = regstats.corr_matrix(df)

    # ----------Build lists------------
    
    title_list = \
    ['Preview','Describe (Numerical)','Describe (Categorical)','Percentile Details',
     'Potential Outliers','Correlation Matrix']
    info_list = \
    [df_preview,df_desc_num, df_desc_cat,percent_values,
     potential_outliers,corr_values]
    error_list = [None,'No numerical data.','All numerical data.','No numerical data.',
                  'No potential outliers.','No categorical, bool, or numerical data.']
    
    # ----------Build output------------
    
    output = ''
    for title, value,error_text in zip(title_list,info_list,error_list):
        if value is None:
            value = "{} skipped: {}".format(title,error_text)
        if str(value).endswith('\n'):
            value = value[:-1]
        output+='{}\n{}\n\n'.format(_io.title_line(title),value)

    # ----------Send to file/print to console------------
    if output_path is None: 
        # Potentially could change this to allow for output_safe to work with directories
         print(output)
    else:
        if not to_folder:
            print('Outputting to file...')
            _io.output_to_file(output,output_path,output_safe)
        else:
            print('Outputting to folder...')
            if not os.path.exists(output_path):
                os.mkdir(output_path)
            for title, value,error_text in zip(title_list,info_list,error_list):
                if value is None:
                    print("{} skipped: {}".format(title,error_text))
                else:
                    file_dir = os.path.join(output_path,"{}.csv".format(title))
                    if type(value) is pd.DataFrame:
                        # Eventually add a check to see if file exists
                        value.to_csv(file_dir)
                    else:
                        _io.output_to_file(value,file_dir,False) 
                        # Change to output_safe when directory output_safe is implemented
        print('Done!')

    # Reset display settings
    pd_settings(*initial_settings)


def pd_settings(max_cols,max_rows,disp_width):
    # Set pandas display settings and retrieve old settings
    initial_max_cols = pd.get_option('display.max_columns')
    initial_max_rows = pd.get_option('display.max_rows')
    initial_width = pd.get_option('display.width')
    
    pd.set_option('display.max_columns', max_cols)
    pd.set_option('display.max_rows', max_rows)
    if disp_width is not None:
        pd.set_option('display.width', disp_width)
    return (initial_max_cols,initial_max_rows,initial_width)

def detailed_desc(df):# Get information from df.describe past
    # The pandas implementation
    try:
        df_desc_num = df.describe(include = 'number').transpose()
    except ValueError:
        df_desc_num = None

    try:
        df_desc_cat = df.describe(exclude = 'number').transpose()
    except ValueError:
        df_desc_cat = None
    return (df_desc_num,df_desc_cat)


# -*- coding: utf-8 -*-
#import pandas as pd
import mr_clean.core.functions.basics as basics
from mr_clean.core.functions.scrub import smart_scrub
from mr_clean.core.functions.colnames import smart_colnames
import pandas as pd
#import mr_clean.core.stats.summary as stats

def clean(df,error_rate = 0):
    """ Superficially cleans data, i.e. changing simple things about formatting.
    Parameters:
    df - DataFrame
        DataFrame to clean
    error_rate - float {0 <= error_rate <= 1}, default 0
        Maximum amount of errors/inconsistencies caused explicitly by cleaning, expressed
        as a percentage of total dataframe rows (0 = 0%, .5 = 50%, etc.)
        Ex: na values from coercing a column of data to numeric
    """
    df = df.copy()

    # Change colnames
    df.columns = smart_colnames(df)
    print('Changed colnames to {}'.format(df.columns))

    # Remove extra whitespace
    obj_col_list = df.select_dtypes(include = 'object').columns
    for col_name in obj_col_list:
        df[col_name] = basics.col_strip(df,col_name)
        print(f"Stripped extra whitespace from '{col_name}'")

    # Coerce columns if possible
    for col_name in obj_col_list:
        new_dtype = coerce_col(df,col_name,error_rate)
        if new_dtype is not None:
            print(f"Coerced '{col_name}' to datatype '{new_dtype}'")

    # Scrub columns
    obj_col_list = df.select_dtypes(include = 'object').columns
    for col_name in obj_col_list:
        scrubf, scrubb = smart_scrub(df,col_name,1-error_rate)
        if scrubf is not None or scrubb is not None:
            print(f"Scrubbed '{scrubf}' from the front and '{scrubb}' from the back of column '{col_name}'")

    # Coerice columns if possible
    for col_name in obj_col_list:
        new_dtype = coerce_col(df,col_name,error_rate)
        if new_dtype is not None:
            print(f"Coerced '{col_name}' to datatype '{new_dtype}'")
    return df

def coerce_col(df,col_name,error_rate):
    """ Change column datatype according to contents of column.
    Parameters:
    df - DataFrame
        DataFrame to edit
    col_name - str
        String that represents a column name
    error_rate - float
        Maximum amount of errors/inconsistencies caused explicitly by this function
    """
    as_cat = df[col_name].astype('category')
    as_num = pd.to_numeric(df[col_name], errors = 'coerce')
    as_dt = pd.to_datetime(df[col_name], errors = 'coerce', infer_datetime_format = True)

    dtypes = ['category','number','datetime']

    new_cols = [as_cat,as_num,as_dt]

    error_rates = [col.isnull().sum() for col in new_cols]
    error_rates[0] = len(new_cols[0].value_counts())
    lowest_rate = min(error_rates)

    for index in range(3):
        if error_rates[index] < error_rate and \
            error_rates[index] <= lowest_rate:
                df[col_name] = new_cols[index]
                return dtypes[index]



# -*- coding: utf-8 -*-
import numpy as np

def row_req(df,cutoff):
    return rows(df) * cutoff

def rows(df):
    return df.shape[0]

def cols(df):
    return df.shape[1]

#def is_num(series): #TODO improve this method
#    if (series.dtype == np.object):
#        return False
#    return True

def get_cutoff(column, cutoff):
    __value_or_container(column,cutoff,1)   

def bc_vec(df,value = True): # Boolean column vector
    return np.ones(rows(df),dtype=bool) if value else np.zeros(rows(df),dtype=bool)

def ic_vec(df,value = 0): # Boolean column vector
    if value == 0:
        return np.zeros(rows(df),dtype=np.int32)
    elif value == 1:
        return np.ones(rows(df),dtype=np.int32)
    else:
        return np.ones(rows(df),dtype=np.int32) * value

def __value_or_container(key,item,default):
    if type(item) is dict:
        return item[key] if key in item else default
    return item

def memory(df, col = None,index = None,deep = True):
    if col is None:
        index = True if index is None else index
        return df.memory_usage(index,deep)
    else:
        index = False if index is None else index
        return df[col].memory_usage(index,deep)


# -*- coding: utf-8 -*-


def substr_unfiltered(suf_tree):
    # Parsing the tree
    # Recursively build suffix tree dictionary
    
    # Turn dictionary into list
    pass

def substr_unfiltered_recurse():
    
    # Using memoization to recursively build dictionary of substrings
    
    pass